#!/usr/bin/env python
# -----------------------------------------------------------------------------
#
# Project: get_min_max.py
# Authors: Stephan Meissl <stephan.meissl@eox.at>
#
# -----------------------------------------------------------------------------
# Copyright (c) 2019 EOX IT Services GmbH
#
# Python script to retrieve min and max values of items.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
# -----------------------------------------------------------------------------


import sys
import os
import argparse
import textwrap
import logging
import traceback
import subprocess
import re

from swiftclient.service import SwiftError, SwiftService


# collection: [name]
COLLECTION_MAP = {
    "VHR_IMAGE_2018": ["VHR IMAGE 2018", ],
}

logger = logging.getLogger(__name__)


def setup_logging(verbosity):
    # start logging setup
    # get command line level
    verbosity = verbosity
    if verbosity == 0:
        level = logging.CRITICAL
    elif verbosity == 1:
        level = logging.ERROR
    elif verbosity == 2:
        level = logging.WARNING
    elif verbosity == 3:
        level = logging.INFO
    else:
        level = logging.DEBUG
    logger.setLevel(level)
    sh = logging.StreamHandler()
    sh.setLevel(level)
    formatter = logging.Formatter("%(asctime)s %(levelname)s: %(message)s")
    sh.setFormatter(formatter)
    logger.addHandler(sh)
    # finished logging setup


def get_min_max(collection):
    logger.info("Starting")

    try:
        with SwiftService() as swift, open("min_max_out", "a+") as outfile:
            p = subprocess.run(
                ["swift", "auth"], capture_output=True,
                timeout=600, check=True
            )
            vars = p.stdout.decode("utf-8")
            os.environ["SWIFT_STORAGE_URL"] = re.findall(
                r"OS_STORAGE_URL=(.*)\n", vars
            )[0]
            os.environ["SWIFT_AUTH_TOKEN"] = re.findall(
                r"OS_AUTH_TOKEN=(.*)\n", vars
            )[0]
            outfile.write("container,product_type,min/max\n")
            try:
                list_gen = swift.list()
                for page in list_gen:
                    if page["success"]:
                        for item in page["listing"]:
                            list_gen2 = swift.list(container=item["name"])
                            for page2 in list_gen2:
                                if page2["success"]:
                                    for item2 in page2["listing"]:
                                        if item2["name"].endswith(".TIF") or \
                                                item2["name"].endswith(".tif"):
                                            gdalout = subprocess.run([
                                                "gdalinfo", "-mm",
                                                "/vsiswift/%s/%s" %
                                                (item["name"], item2["name"])],
                                                capture_output=True,
                                                timeout=600, check=True
                                            ).stdout.decode("utf-8")
                                            minmax = re.findall(
                                                r"Computed Min/Max=(.*)\n",
                                                gdalout
                                            )
                                            outfile.write(
                                                "%s,%s,%s\n" %
                                                (item["name"],
                                                 item2["name"].split("/")[1],
                                                 minmax)
                                            )
                                else:
                                    logger.error(
                                        "No product found in container '%s'."
                                        % item["name"]
                                    )
                                    return(1)
                    else:
                        logger.error("No container found.")
                        return(1)

            except SwiftError as e:
                logger.debug(traceback.format_exc())
                logger.error("%s: %s\n" % (type(e).__name__, str(e)))
                return(1)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.error("%s: %s\n" % (type(e).__name__, str(e)))
        return(1)

    logger.info("Successfully finished")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.description = textwrap.dedent("""\
    Get min/max of products.
    """)

    parser.add_argument(
        "collection", default=None,
        help=(
            "Collection the registrar is run for."
        )
    )

    parser.add_argument(
        "-v", "--verbosity", type=int, default=3, choices=[0, 1, 2, 3, 4],
        help=(
            "Set verbosity of log output "
            "(4=DEBUG, 3=INFO, 2=WARNING, 1=ERROR, 0=CRITICAL). (default: 3)"
        )
    )

    arg_values = parser.parse_args()

    setup_logging(arg_values.verbosity)

    collection = arg_values.collection
    if collection not in COLLECTION_MAP:
        logger.critical("Provided collection '%s' is not valid." % collection)
        sys.exit(1)

    get_min_max(
        collection,
    )
